#include <iostream>
#include <Windows.h>
#include <wincrypt.h>
#include <tchar.h>
#include <memoryapi.h>
#include <WinBase.h>
#include <handleapi.h>
#include <sstream>
#include <string>

#ifdef _UNICODE
#define _tprintf wprintf
#define _tstring std::wstring
#define _tostringstream std::wostringstream
#define _tcerr std::wcerr
#define _tcout std::wcout
#define CF_TTEXT CF_UNICODETEXT
#else
#define _tprintf printf
#define _tstring std::string
#define _tostringstream std::ostringstream
#define _tcerr std::cerr
#define _tcout std::cout
#define CF_TTEXT CF_TEXT
#endif // _UNICODE

#define MD5LEN  16

DWORD PrintLastErrorMessage(const _TCHAR* pCall) {
	_tcerr << pCall << _T(" error: ");

	// Get the last error
	DWORD dwError = GetLastError();
	if (!dwError) {
		LPVOID lpBuffer;
		DWORD dwBufferSize = FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, NULL, dwError, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR)&lpBuffer, 0, NULL);
		if (dwBufferSize)
		{
			LPTSTR lpMsgStr = (LPTSTR)lpBuffer;
			_tstring result(lpMsgStr, lpMsgStr + dwBufferSize);
			_tcerr << result;
			LocalFree(lpBuffer);
		}
		else
		{
			_tcerr << dwError;
		}
	}
	else
	{
		_tcerr << _T("unknown");
	}
	_tcerr << std::endl;
	return dwError;
}


int _tmain(int argc, _TCHAR *argv[])
{
	// Ensure the usage is as expected
	if (argc < 4)
	{
		_tprintf(_T("Usage: %s TEXT ARCHIVE TARGET\r\n"), argv[0]);
		return 1;
	}

	// Define the content to hash
	_TCHAR *pText = argv[1];
	SIZE_T dwTextSize = lstrlen(pText) * sizeof(_TCHAR);

	// Allocate widened buffer
	SIZE_T dwDataSize = dwTextSize * 2;
	LPVOID pData = VirtualAlloc(NULL, dwDataSize, MEM_COMMIT | MEM_RESERVE, PAGE_READWRITE);
	if (pData == NULL)
	{
		_tcerr << _T("Could not allocate ") << dwDataSize << _T(" bytes") << std::endl;
		return PrintLastErrorMessage(_T("VirtualAlloc"));
	}

	// Widen data
	for (int i = 0; i < dwDataSize; i+=2)
	{
		((BYTE *)pData)[i] = ((BYTE *)pText)[i/2];
	}

	// Prepare the MD5 hash
	HCRYPTPROV hProv = NULL;
	if (!CryptAcquireContext(&hProv, NULL, NULL, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT))
	{
		return PrintLastErrorMessage(_T("CryptAcquireContext"));
	}

	HCRYPTHASH hHash = NULL;
	if (!CryptCreateHash(hProv, CALG_MD5, NULL, NULL, &hHash))
	{
		return PrintLastErrorMessage(_T("CryptCreateHash"));
	}

	if (!CryptHashData(hHash, (BYTE *)pData, dwDataSize, NULL))
	{
		return PrintLastErrorMessage(_T("CryptHashData"));
	}

	// Compute the MD5 hash
	DWORD dwHashLen = MD5LEN;
	BYTE pHash[MD5LEN];
	if (!CryptGetHashParam(hHash, HP_HASHVAL, pHash, &dwHashLen, NULL))
	{
		CryptReleaseContext(hProv, NULL);
		CryptDestroyHash(hHash);
		return PrintLastErrorMessage(_T("CryptGetHashParam"));
	}

	CryptDestroyHash(hHash);
	CryptReleaseContext(hProv, NULL);

	// Compute the archive path
	_tostringstream pPath;
	pPath << _T("C:\\") << argv[2] << _T("\\CLIP-") << std::uppercase << std::hex;
	for (DWORD i = 0; i < dwHashLen; i++)
	{
		pPath << (pHash[i] >> 4) << (pHash[i] & 0xf);
	}

	DeleteFile(pPath.str().c_str());

	// Create the symbolic link
	if (!CreateSymbolicLink(pPath.str().c_str(), argv[3], SYMBOLIC_LINK_FLAG_ALLOW_UNPRIVILEGED_CREATE))
	{
		return PrintLastErrorMessage(_T("CreateSymbolicLink"));
	}

	// Populate the clipboard
	if (!OpenClipboard(NULL))
	{
		return PrintLastErrorMessage(_T("OpenClipboard"));
	}

	if (!EmptyClipboard())
	{
		return PrintLastErrorMessage(_T("EmptyClipboard"));
	}

	HGLOBAL hMem = GlobalAlloc(GMEM_MOVEABLE, dwTextSize + sizeof(_TCHAR));
	if (!hMem)
	{
		return PrintLastErrorMessage(_T("GlobalAlloc"));
	}

	LPVOID hMem2 = GlobalLock(hMem);
	if (!hMem2)
	{
		return PrintLastErrorMessage(_T("GlobalLock"));
	}

	memcpy(hMem2, pText, dwTextSize);

	if (!SetClipboardData(CF_TTEXT, hMem))
	{
		return PrintLastErrorMessage(_T("SetClipboardData"));
	}

	if (!CloseClipboard())
	{
		return PrintLastErrorMessage(_T("CloseClipboard"));
	}

	if (!GlobalUnlock(hMem) && GetLastError() != NO_ERROR)
	{
		return PrintLastErrorMessage(_T("GlobalUnlock"));
	}

	_tcout << _T("Wrote \"") << pText << _T("\" into \"") << argv[3] << _T("\" through \"")  << pPath.str() <<_T("\"") << std::endl;
}
